"""Converts Python types to string representations suitable for Maps API server.
    For example:
    sydney = {
        "lat" : -33.8674869,
        "lng" : 151.2069902
    }
    convert.latlng(sydney)
    # '-33.8674869,151.2069902'
"""

import time as _time


def format_float(arg):
    """Formats a float value to be as short as possible.
    Trims extraneous trailing zeros and period to give API
    args the best possible chance of fitting within 2000 char
    URL length restrictions.
    For example:
    format_float(40) -> "40"
    format_float(40.0) -> "40"
    format_float(40.1) -> "40.1"
    format_float(40.001) -> "40.001"
    format_float(40.0010) -> "40.001"
    :param arg: The lat or lng float.
    :type arg: float
    :rtype: string
    """
    return ("%f" % float(arg)).rstrip("0").rstrip(".")


def latlng(arg):
    """Converts a lat/lon pair to a comma-separated string.
    For example:
    sydney = {
        "lat" : -33.8674869,
        "lng" : 151.2069902
    }
    convert.latlng(sydney)
    # '-33.8674869,151.2069902'
    For convenience, also accepts lat/lon pair as a string, in
    which case it's returned unchanged.
    :param arg: The lat/lon pair.
    :type arg: string or dict or list or tuple
    """
    if is_string(arg):
        return arg

    normalized = normalize_lat_lng(arg)
    return "%s,%s" % (format_float(normalized[0]), format_float(normalized[1]))


def normalize_lat_lng(arg):
    """Take the various lat/lng representations and return a tuple.
    Accepts various representations:
    1) dict with two entries - "lat" and "lng"
    2) list or tuple - e.g. (-33, 151) or [-33, 151]
    :param arg: The lat/lng pair.
    :type arg: dict or list or tuple
    :rtype: tuple (lat, lng)
    """
    if isinstance(arg, dict):
        if "lat" in arg and "lng" in arg:
            return arg["lat"], arg["lng"]
        if "latitude" in arg and "longitude" in arg:
            return arg["latitude"], arg["longitude"]

    # List or tuple.
    if _is_list(arg):
        return arg[0], arg[1]

    raise TypeError(
        "Expected a lat/lng dict or tuple, "
        "but got %s" % type(arg).__name__)


def location_list(arg):
    """Joins a list of locations into a pipe separated string, handling
    the various formats supported for lat/lng values.
    For example:
    p = [{"lat" : -33.867486, "lng" : 151.206990}, "Sydney"]
    convert.waypoint(p)
    # '-33.867486,151.206990|Sydney'
    :param arg: The lat/lng list.
    :type arg: list
    :rtype: string
    """
    if isinstance(arg, tuple):
        # Handle the single-tuple lat/lng case.
        return latlng(arg)
    else:
        return "|".join([latlng(location) for location in as_list(arg)])


def join_list(sep, arg):
    """If arg is list-like, then joins it with sep.
    :param sep: Separator string.
    :type sep: string
    :param arg: Value to coerce into a list.
    :type arg: string or list of strings
    :rtype: string
    """
    return sep.join(as_list(arg))


def as_list(arg):
    """Coerces arg into a list. If arg is already list-like, returns arg.
    Otherwise, returns a one-element list containing arg.
    :rtype: list
    """
    if _is_list(arg):
        return arg
    return [arg]


def _is_list(arg):
    """Checks if arg is list-like. This excludes strings and dicts."""
    if isinstance(arg, dict):
        return False
    if isinstance(arg, str):  # Python 3-only, as str has __iter__
        return False
    return (not _has_method(arg, "strip")
            and _has_method(arg, "__getitem__")
            or _has_method(arg, "__iter__"))


def is_string(val):
    """Determines whether the passed value is a string, safe for 2/3."""
    try:
        basestring
    except NameError:
        return isinstance(val, str)
    return isinstance(val, basestring)


def time(arg):
    """Converts the value into a unix time (seconds since unix epoch).
    For example:
        convert.time(datetime.now())
        # '1409810596'
    :param arg: The time.
    :type arg: datetime.datetime or int
    """
    # handle datetime instances.
    if _has_method(arg, "timetuple"):
        arg = _time.mktime(arg.timetuple())

    if isinstance(arg, float):
        arg = int(arg)

    return str(arg)


def _has_method(arg, method):
    """Returns true if the given object has a method with the given name.
    :param arg: the object
    :param method: the method name
    :type method: string
    :rtype: bool
    """
    return hasattr(arg, method) and callable(getattr(arg, method))


def components(arg):
    """Converts a dict of components to the format expected by the Google Maps
    server.
    For example:
    c = {"country": "US", "postal_code": "94043"}
    convert.components(c)
    # 'country:US|postal_code:94043'
    :param arg: The component filter.
    :type arg: dict
    :rtype: basestring
    """

    # Components may have multiple values per type, here we
    # expand them into individual key/value items, eg:
    # {"country": ["US", "AU"], "foo": 1} -> "country:AU", "country:US", "foo:1"
    def expand(arg):
        for k, v in arg.items():
            for item in as_list(v):
                yield "%s:%s" % (k, item)

    if isinstance(arg, dict):
        return "|".join(sorted(expand(arg)))

    raise TypeError(
        "Expected a dict for components, "
        "but got %s" % type(arg).__name__)


def bounds(arg):
    """Converts a lat/lon bounds to a comma- and pipe-separated string.
    Accepts two representations:
    1) string: pipe-separated pair of comma-separated lat/lon pairs.
    2) dict with two entries - "southwest" and "northeast". See convert.latlng
    for information on how these can be represented.
    For example:
    sydney_bounds = {
        "northeast" : {
            "lat" : -33.4245981,
            "lng" : 151.3426361
        },
        "southwest" : {
            "lat" : -34.1692489,
            "lng" : 150.502229
        }
    }
    convert.bounds(sydney_bounds)
    # '-34.169249,150.502229|-33.424598,151.342636'
    :param arg: The bounds.
    :type arg: dict
    """

    if isinstance(arg, dict):
        if "southwest" in arg and "northeast" in arg:
            return "%s|%s" % (latlng(arg["southwest"]),
                              latlng(arg["northeast"]))

    raise TypeError(
        "Expected a bounds (southwest/northeast) dict, "
        "but got %s" % type(arg).__name__)


def decode_polyline(polyline):
    """Decodes a Polyline string into a list of lat/lng dicts.
    See the developer docs for a detailed description of this encoding:
    https://developers.google.com/maps/documentation/utilities/polylinealgorithm
    :param polyline: An encoded polyline
    :type polyline: string
    :rtype: list of dicts with lat/lng keys
    """
    points = []
    index = lat = lng = 0

    while index < len(polyline):
        result = 1
        shift = 0
        while True:
            b = ord(polyline[index]) - 63 - 1
            index += 1
            result += b << shift
            shift += 5
            if b < 0x1f:
                break
        lat += (~result >> 1) if (result & 1) != 0 else (result >> 1)

        result = 1
        shift = 0
        while True:
            b = ord(polyline[index]) - 63 - 1
            index += 1
            result += b << shift
            shift += 5
            if b < 0x1f:
                break
        lng += ~(result >> 1) if (result & 1) != 0 else (result >> 1)

        points.append({"lat": lat * 1e-5, "lng": lng * 1e-5})

    return points


def encode_polyline(points):
    """Encodes a list of points into a polyline string.
    See the developer docs for a detailed description of this encoding:
    https://developers.google.com/maps/documentation/utilities/polylinealgorithm
    :param points: a list of lat/lng pairs
    :type points: list of dicts or tuples
    :rtype: string
    """
    last_lat = last_lng = 0
    result = ""

    for point in points:
        ll = normalize_lat_lng(point)
        lat = int(round(ll[0] * 1e5))
        lng = int(round(ll[1] * 1e5))
        d_lat = lat - last_lat
        d_lng = lng - last_lng

        for v in [d_lat, d_lng]:
            v = ~(v << 1) if v < 0 else v << 1
            while v >= 0x20:
                result += (chr((0x20 | (v & 0x1f)) + 63))
                v >>= 5
            result += (chr(v + 63))

        last_lat = lat
        last_lng = lng

    return result


def shortest_path(locations):
    """Returns the shortest representation of the given locations.
    The Elevations API limits requests to 2000 characters, and accepts
    multiple locations either as pipe-delimited lat/lng values, or
    an encoded polyline, so we determine which is shortest and use it.
    :param locations: The lat/lng list.
    :type locations: list
    :rtype: string
    """
    if isinstance(locations, tuple):
        # Handle the single-tuple lat/lng case.
        locations = [locations]
    encoded = "enc:%s" % encode_polyline(locations)
    unencoded = location_list(locations)
    if len(encoded) < len(unencoded):
        return encoded
    else:
        return unencoded



"""
Entry point for raw requests. Can be used to compute queries of API parts for which specific methods have not been yet implemented, and still benefit from the core methods.

for instance, client.raw()
"""


def raw(client, url, extra_params=None, verbose=False):
    return client._get(url=url, extra_params=extra_params, verbose=verbose)



"""
This module computes API requests.
"""
import functools
import os
from datetime import datetime, timedelta
import time

import json
import requests

from multiprocessing import Pool
from navitia_client.utils import important_print
import navitia_client

_RETRIABLE_STATUSES = set([500, 503, 504])


class Client(object):
    """
    Performs requests to the navitia API web services.
    """

    def __init__(self, user, retry_timeout=20, core_url='sncf'):
        if core_url == 'sncf':
            self.core_url = "https://api.sncf.com/v1/"
        else:
            self.core_url = core_url
        self.user = user
        self.retry_timeout = retry_timeout
        self.last_url = None

    def _get(self, url, extra_params=None, verbose=False, first_request_time=None, retry_counter=0, ignore_fail=False):
        if verbose and not first_request_time:
            print("Import on url %s " % url)

        if not first_request_time:
            first_request_time = datetime.now()

        elapsed = datetime.now() - first_request_time
        if elapsed > timedelta(seconds=self.retry_timeout):
            raise navitia_client.exceptions.Timeout()

        if retry_counter > 0:
            # 0.5 * (1.5 ^ i) is an increased sleep time of 1.5x per iteration,
            # starting at 0.5s when retry_counter=0. The first retry will occur
            # at 1, so subtract that first.
            delay_seconds = 0.5 * 1.5 ** (retry_counter - 1)
            time.sleep(delay_seconds)

        full_url = os.path.join(self.core_url, url)

        try:
            response = requests.get(
                url=full_url, auth=(self.user, ""), params=(extra_params or {}))

        except requests.exceptions.Timeout:
            if not ignore_fail:
                raise navitia_client.exceptions.Timeout()
            else:
                return False
        except Exception as e:
            if not ignore_fail:
                raise navitia_client.exceptions.TransportError(e)
            else:
                return False

        # Warn if not 200
        if response.status_code != 200:
            print("WARNING: response status_code is %s" % response.status_code)

        if response.status_code in _RETRIABLE_STATUSES:
            # Retry request.
            print("WARNING: retry number %d" % retry_counter)
            return self._get(url=url, extra_params=extra_params, first_request_time=first_request_time, retry_counter=retry_counter + 1, verbose=verbose, ignore_fail=ignore_fail)

        return response

    def _get_single_page_multiprocess(self, url, page, count, extra_params=None, verbose=False):
        """
        Return tuple: (page_number,response)
        Needed to perform multiprocessed requests.
        """
        if verbose:
            print("Import on page %d " % page)
        # Pagination parameters of request
        pagination_params = {
            "start_page": page,
            "count": count,
        }
        parameters = {**pagination_params, **(extra_params or {})}
        # Ignore fail: if one request fails, we still want others to be
        # computed
        response = self._get(
            url=url, extra_params=parameters, ignore_fail=True)

        return (page, response)

    def _extract_nbr_results(self, response):
        """
        Out of a request response, finds total number of results
        """
        parsed = json.loads(response.text)
        try:
            nbr_results = parsed["pagination"]["total_result"]
            return nbr_results
        except KeyError:
            # No pagination in page
            print("WARNING: not able to extract pagination out of first request.")
            return False


def make_api_method(func):
    """
    Provides a single entry point for modifying all API methods.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result
    return wrapper

from navitia_client.journeys import journeys
from navitia_client.route_schedules import route_schedules
from navitia_client.multipage import multipage
from navitia_client.raw import raw

Client.journeys = make_api_method(journeys)
Client.route_schedules = make_api_method(route_schedules)
Client.multipage = make_api_method(multipage)
Client.raw = make_api_method(raw)



"""Performs multi-pages requests to the Navitia journeys API.

The requests are parallelized to reduce significantly the request time.

Most useful with:
- connections (walking connections between stop_points?)
- disruptions (trip disruptions, one-to-one relationship with trips)
- journey_pattern_points
- journey_patterns
- lines
- routes
- stop_areas (area of aggregation of stop_points)
- stop_points (where journeys can stop)
- trips
- vehicle_journeys
"""

from navitia_client.client import Client
from navitia_client.utils import important_print
from multiprocessing import Pool
from datetime import datetime


def unwrap_self(arg, **kwarg):
    """ To unwrap self in multiprocessing mode
    http://www.rueckstiess.net/research/snippets/show/ca1d7d90
    """
    return Client._get_single_page_multiprocess(*arg, **kwarg)


def multipage(client, url, page_limit=10, count=100, extra_params=None, verbose=False, log=False):
    """
    url: str
    #

    page_limit: int
    #

    count: int
    #

    extra_params: dict
    #

    verbose: boolean
    #

    log: boolean
    #

    ############
    return: dict
        key: page number
        value: response object
    """
    responses = {}

    # First page
    page, response = client._get_single_page_multiprocess(
        url=url, page=0, count=count, verbose=verbose, extra_params=extra_params)
    responses[page] = response

    # Find number of requests to make
    nbr_results = client._extract_nbr_results(responses[0])
    if not nbr_results:
        # If unable to get pagination count, returns first response
        return responses

    # Find number of pages to query
    blocs = nbr_results // count
    page_limit = min(page_limit, blocs + 1)
    if verbose:
        important_print("There are %d elements with %d elements per page. Limit is %d." % (
            nbr_results, count, page_limit))

    # Query other pages, multiprocessing
    pages = range(1, page_limit)
    pool = Pool(processes=30)
    n = len(pages)
    all_parameters = zip([client] * n, [url] * n,
                         pages, [count] * n, [extra_params or {}] * n, [verbose] * n)

    list_tuples = pool.map(unwrap_self, all_parameters)
    pool.close()
    pool.join()
    # Add results to initial results (first page)
    responses.update(dict(list_tuples).items())

    # Save logs
    if log:
        log = {}
        log["date"] = str(datetime.now().isoformat())
        log["url"] = url
        log["count_per_page"] = count
        log["page_limit"] = page_limit
        log["nbr_results"] = nbr_results
        responses["log"] = log

    return responses



"""Performs requests to the Navitia journeys API.
http://doc.navitia.io/#journeys

Two accesses:
/coverage/{region_id}/journeys 	List of journeys on a specific coverage
/coverage/{a_path_to_resource}/journeys

Focus on first access:
# from/to parameters
##coordonates: lon;lat
- coverage/sncf/journeys?from=2.3749036;48.8467927&to=2.2922926;48.8583736&
- coverage/sncf/journeys?from=2.3749036%3B48.8467927&to=2.2922926%3B48.8583736&
##ressources id:
- coverage/sncf/journeys?from=stop_point:OCE:SP:CorailIntercit%C3%A9-87113001
- coverage/sncf/journeys?from=stop_area:OCE:SA:87171009
# datetime parameter:
&datetime=20161221T000000
"""


def journeys(client, origin=None, destination=None, datetime=None, datetime_represents=None, data_freshness=None, verbose=False):
    """
    from:
    # The id of the departure of your journey. If none are provided an isochrone is computed, can be:
        - a ressource id:
        - coordonates, lon;lat:


    to:
    # The id of the arrival of your journey. If none are provided an isochrone is computed

    datetime:
    # Date and time to go: iso-date-time

    datetime_represents:
    # Can be departure or arrival.
    # If departure, the request will retrieve journeys starting after datetime.
    # If arrival it will retrieve journeys arriving before datetime.

    data_freshness:
    #Define the freshness of data to use to compute journeys
        - realtime
        - base_schedule
    when using the following parameter “&data_freshness=base_schedule”
    you can get disrupted journeys in the response. You can then display the disruption message to the traveler and make a realtime request to get a new undisrupted solution.

    """

    # checks if good format for from and to

    params = {}

    if not origin and not destination:
        raise ValueError("Must specify at least origin or destination")

    if origin:
        params["from"] = origin

    if destination:
        params["to"] = destination

    # checks if datetime format is good

    if datetime:
        params["datetime"] = datetime

    if datetime_represents in [None, "departure", "arrival"]:
        params["datetime_represents"] = datetime_represents
    else:
        raise ValueError(
            "datetime_represents if specified must be 'departure' or 'arrival'")

    if data_freshness in [None, "realtime", "base_schedule"]:
        params["data_freshness"] = data_freshness
    else:
        raise ValueError(
            "Data freshness if specified must be 'realtime' or 'base_schedule'")

    return client._get(url="coverage/sncf/journeys", extra_params=params, verbose=verbose)



"""
Defines exceptions that are thrown by the Navitia client.
"""


class ApiError(Exception):
    """Represents an exception returned by the remote API."""

    def __init__(self, status, message=None):
        self.status = status
        self.message = message

    def __str__(self):
        if self.message is None:
            return self.status
        else:
            return "%s (%s)" % (self.status, self.message)


class TransportError(Exception):
    """Something went wrong while trying to execute the request."""

    def __init__(self, base_exception=None):
        self.base_exception = base_exception

    def __str__(self):
        if self.base_exception:
            return str(self.base_exception)

        return "An unknown error occurred."


class HTTPError(TransportError):
    """An unexpected HTTP error occurred."""

    def __init__(self, status_code):
        self.status_code = status_code

    def __str__(self):
        return "HTTP Error: %d" % self.status_code


class Timeout(Exception):
    """The request timed out."""
    pass


class _RetriableRequest(Exception):
    """Signifies that the request can be retried."""
pass



# http://doc.navitia.io/#route-schedules
"""
Also known as /route_schedules service.

This endpoint gives you access to schedules of routes (so a kind of time table), with a response made of an array of route_schedule, and another one of note. You can access it via that kind of url: https://api.navitia.io/v1/{a_path_to_a_resource}/route_schedules

URLs:
/coverage/{region_id}/{resource_path}/route_schedules

trips:
/coverage/sncf/trips/{trip_id}/route_schedules
/coverage/{lon;lat}/coords/{lon;lat}/route_schedules

stop_points
/coverage/sncf/stop_points/stop_point:OCE:SP:CorailIntercit%C3%A9-87116137/route_schedules

routes
/coverage/sncf/routes/route:OCE:1-TrainTER-87471003-87474007/route_schedules

lines
coverage/sncf/lines/line:OCE:SN-87775866-87775007/route_schedules
"""

import os


def route_schedules(client, raw=None, line=None, route=None, trip=None, stop_point=None, from_datetime=None, duration=None, items_per_schedule=None, data_freshness=None, disable_geojson=None, verbose=False):
    """
    from_datetime: iso-date-time
    # The date_time from which you want the schedules

    duration : int
    # Maximum duration in seconds between from_datetime and the retrieved datetimes. (default in api: 86400)

    items_per_schedule: int
    # Maximum number of columns per schedule.

    data_freshness: enum
    # Define the freshness of data to use
        - realtime
        - base_schedule (default)

        disable_geojson: boolean
    # Default false

    """
    params = {}

    url_begin = "coverage/sncf/"
    url_end = "route_schedules"

    # Maximum one, out of: trip_id, stop_point, route, line, raw
    if raw:
        # /coverage/sncf/routes/route:OCE:1-TrainTER-87471003-87474007/route_schedules
        # raw = "routes/route:OCE:1-TrainTER-87471003-87474007/"
        # made for all use cases
        full_url = os.path.join(url_begin, raw, url_end)

    if route:
        # /coverage/sncf/routes/route:OCE:1-TrainTER-87471003-87474007/route_schedules
        full_url = os.path.join(url_begin, "routes", route, url_end)

    if trip:
        # /coverage/sncf/trips/{trip_id}/route_schedules
        # check trip_id format?
        full_url = os.path.join(url_begin, "trips", trip, url_end)

    if stop_point:
        # /coverage/sncf/stop_points/stop_point:OCE:SP:CorailIntercit%C3%A9-87116137/route_schedules
        # check id format?
        full_url = os.path.join(url_begin, "stop_points", stop_point, url_end)

    if line:
        # coverage/sncf/lines/line:OCE:SN-87775866-87775007/route_schedules
        # check id format?
        full_url = os.path.join(url_begin, "lines", line, url_end)

    if from_datetime:
        # TODO: check is well formated
        params["from_datetime"] = from_datetime

    if duration:
        if isinstance(duration, int):
            params["duration"] = duration
        else:
            raise ValueError("Duration should be integer")

    if data_freshness in [None, "realtime", "base_schedule"]:
        params["data_freshness"] = data_freshness
    else:
        raise ValueError(
            "Data freshness if specified must be 'realtime' or 'base_schedule'")

    if isinstance(items_per_schedule, int):
        params["items_per_schedule"] = items_per_schedule

    if disable_geojson:
        params["disable_geojson"] = True

    return client._get(url=full_url, extra_params=params, verbose=verbose)



"""
This module contains some useful functions.

For now:
 - flatten_columns
 - flatten_dataframe
"""


def important_print(message, level=0):
    if level == 1:
        print("-" * 70)
        print(message.capitalize())
        print("-" * 70)
    if level == 0:
        print("-" * 50)
        print(message)
        print("-" * 50)


def flattenjson(b, delim):
    val = {}
    if isinstance(b, dict):
        for i in b.keys():
            if isinstance(b[i], dict):
                get = flattenjson(b[i], delim)
                for j in get.keys():
                    val[i + delim + j] = get[j]
            else:
                val[i] = b[i]
    return val


def length_with_nan(x):
    if isinstance(x, list):
        return len(x)
    else:
        return x


def first_with_nan(x):
    if isinstance(x, list):
        return x[0]
    else:
        return x


def keys_with_nan(x):
    if isinstance(x, dict):
        return str(list(x.keys()))
    else:
        return x


def check_asked_cols(df_to_check, cols_to_check):
    real_columns = df_to_check.columns
    for column in cols_to_check:
        if column not in real_columns:
            cols_to_check.remove(column)
    return cols_to_check


def flatten_columns(df, columns_list, drop=False, debug=False):
    """
    Goal: flatten a dataframe, which has columns filled with lists or dictionnaries.

    Parameters:
    - df : dataframe: the dataframe you want to flatten
    - columns_list: list of strings: the columns you want to flatten
    - drop: boolean: to drop flattened columns

    Dataframe is changed in place, with new columns, and flattened columns deleted if drop=True.
    Returns added columns
    """

    # First, we check if asked columns are present
    columns_list = check_asked_cols(df, columns_list)

    # Instanciate list of created columns during process
    added_columns = []
    for column_flattened in columns_list:
        if debug:
            print("-" * 20)
            print("FLATTENING " + column_flattened.capitalize())

        # Find out whether the value is a dict
        # TODO check on whole column
        if isinstance(df[column_flattened][0], dict):

            # Lets check if keys are all the same:
            # First we find keys on the dictionnary
            series_keylists = df[column_flattened].apply(keys_with_nan)
            # We count occurences of unique lists of keys
            keylists_counts = series_keylists.value_counts()

            # If there is always the same key we can flatten
            if len(keylists_counts) == 1:
                if debug:
                    print("This columns has dicts of same keys. Ok to flatten.")
                keys_to_flatten = list(df[column_flattened][0].keys())
                for key in keys_to_flatten:

                    def key_with_nan(x):
                        try:
                            return x[key]
                        except:
                            return x

                    df[column_flattened + "_" +
                        key] = df[column_flattened].apply(key_with_nan)
                    added_columns.append(column_flattened + "_" + key)
                    if debug:
                        print("Created column : " +
                              column_flattened + "_" + key)
                if drop:
                    df.drop(column_flattened, axis=1, inplace=True)
                    if debug:
                        print("Removed original " + column_flattened)
            else:
                if debug:
                    print("Keys are not the same on all rows.")
                    print(keylists_counts)

        # Find out whether the value is a list
        # TODO check on whole column
        elif isinstance(df[column_flattened][0], list):
            if debug:
                print("Column " + column_flattened + " is a list.")
            # check if all of same size

            series_sizes = df[column_flattened].apply(length_with_nan)

            sizes_counts = series_sizes.value_counts()
            if debug:
                print("Counts of different sizes: ", len(sizes_counts))
            # if all of same size
            if len(sizes_counts) == 1:
                if debug:
                    print("All of same size: " + str(sizes_counts))
                # and if size is one: we can flatten list [x] -> x
                if series_sizes[0] == 1:
                    if debug:
                        print(
                            "All lists are from size one, so we can take flatten it.")
                    df[column_flattened] = df[
                        column_flattened].apply(first_with_nan)
                    added_columns.append(column_flattened)
                    if debug:
                        print("Replaced column " + column_flattened)
            else:
                if debug:
                    print(
                        "Size are not the same or >1 so we cannot flatten list, but we can count the number of elements.")
                df[column_flattened +
                    "_size"] = df[column_flattened].apply(length_with_nan)
                added_columns.append(column_flattened + "_size")
                if debug:
                    print("New column: " + column_flattened + "_size")
        else:
            if debug:
                print("Column " + column_flattened +
                      " is not a dictionnary nor a list!")
    return added_columns


def flatten_dataframe(df, drop=False, max_depth=3, debug=False):
    """
    Flatten all columns of a given dataframe, with a max_depth defined.
    """
    cols_to_flatten = df.columns
    cols_flattened = []
    k = 1
    while k <= max_depth:
        if debug:
            print("-" * 30)
            print("FLATENNING LEVEL " + str(k))
            print("-" * 30)
        # we use new columns to flatten them
        cols_to_flatten = flatten_columns(df, cols_to_flatten, drop)
        cols_flattened.append(cols_to_flatten)
        k += 1
        if len(cols_to_flatten) == 0:
            if debug:
                print("-" * 30)
                print("END NO MORE COLUMNS TO FLATTEN")
                print("-" * 30)
            break
    return cols_flattened



import os
import json
import pandas as pd
from navitia_client.utils import flatten_dataframe, important_print


class Parser:

    def __init__(self, request_results, asked_path):
        self.asked_path = asked_path
        self.results = request_results
        self.item_name = os.path.basename(asked_path)
        self.parsed = {}  # dictionary of page : dictionary
        self.parsing_errors = {}
        self.nested_items = None  # will be a dict
        self.unnested_items = None  # will be a dict
        self.links = []  # first page is enough
        self.disruptions = {}  # all pages
        self.keys = []  # collect keys found in request answer
        self.nbr_expected_items = None
        self.nbr_collected_items = None
        self.log = None

    def set_results(self, request_results):
        self.results = request_results

    def parse(self):
        self.parse_requests()
        self.extract_keys()
        self.extract_links()
        self.extract_disruptions()
        self.get_nested_items()
        self.get_unnested_items()
        self.extract_nbr_expected_items()
        self.count_nbr_collected_items()
        self.parse_log()

    def parse_requests(self):
        # First operation, to parse requests text into python dictionnaries
        for page, value in self.results.items():
            # Only add if answer was good
            if value.status_code == 200:
                try:
                    self.parsed[page] = json.loads(value.text)
                except ValueError:
                    print("JSON decoding error.")
                    self.parsing_errors[page] = "JSON decoding error"

    def get_nested_items(self):
        """
        Result is a dictionary, of one key: item_name, and value is list of items (concatenate all result pages).
        """
        dictionnary = {self.item_name: []}
        for page, value in self.parsed.items():
            # concatenate two lists of items
            dictionnary[self.item_name] += value[self.item_name]
        self.nested_items = dictionnary

    def get_nested_disruptions(self):
        """
        Result is a dictionary, of one key: item_name, and value is list of items (concatenate all result pages).
        """
        if self.item_name == "disruptions":
            return False

        dictionnary = {"disruptions": []}
        for page, value in self.parsed.items():
            # concatenate two lists of items
            dictionnary["disruptions"] += value["disruptions"]
        self.disruptions = dictionnary

    def get_unnested_items(self):
        df = pd.DataFrame(self.nested_items[self.item_name])
        flatten_dataframe(df, drop=True, max_depth=5)
        self.unnested_items = df.to_dict()

    def extract_keys(self):
        # Extract keys of first page
        self.keys = list(self.parsed[0].keys())

    def extract_links(self):
        # Extract from first page
        try:
            self.links = self.parsed[0]["links"]
        except KeyError:
            self.links = {"links": "Not found"}

    def extract_disruptions(self):
        # TODO extract all pages
        # Extract from first page
        try:
            self.disruptions = self.parsed[0]["disruptions"]
        except KeyError:
            self.disruptions = {"disruptions": "Not found"}

    def extract_nbr_expected_items(self):
        if self.results[0].status_code != 200:
            return None
        # Parse first request answer.
        parsed = json.loads(self.results[0].text)
        # Extract pagination part.
        pagination = parsed["pagination"]
        # Extract total_result
        self.nbr_expected_items = pagination["total_result"]

    def count_nbr_collected_items(self):
        unnested = pd.DataFrame(self.unnested_items)  # df
        self.nbr_collected_items = len(unnested.index)

    def explain(self):
        print("Parsing:")
        print("Keys found: " + str(self.keys))
        print(self.item_name.capitalize() + " has " +
              str(self.nbr_expected_items) + " elements.")

    def parse_log(self):
        log = {}
        log["number_requests"] = len(self.results)
        log["number_parsed"] = len(self.parsed)
        log["keys"] = self.keys
        log["nbr_announced_items"] = self.nbr_expected_items
        log["nbr_collected_items"] = self.nbr_collected_items
        log["item_columns"] = list(pd.DataFrame(
            self.unnested_items).columns.values)
        self.log = log
        log["parsing_errors"] = self.parsing_errors

    def write_all(self, directory):
        # Get results
        unnested = pd.DataFrame(self.unnested_items)  # df
        nested = self.nested_items  # dict
        # Write item csv
        unnested.to_csv(os.path.join(directory, self.item_name + ".csv"))
        # Write item json
        with open(os.path.join(directory, self.item_name + ".json"), 'w') as f:
            json.dump(nested, f, ensure_ascii=False)
        # Write links (of first page)
        with open(os.path.join(directory, "links.json"), 'w') as f:
            json.dump(self.links, f, ensure_ascii=False)
        # Write disruptions (if item different)
        if self.item_name != "disruptions":
            unnested_dis = pd.DataFrame(self.disruptions)  # df
            unnested_dis.to_csv(os.path.join(directory, "disruptions.csv"))
        # Write logs
        with open(os.path.join(directory, "parse_log.json"), 'w') as f:
            json.dump(self.log, f, ensure_ascii=False)



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='navitia_client',
    packages=['navitia_client'],  # this must be the same as the name above
    version='0.1',
    description='(unofficial) navitia client for python',
    author='Leonard Binet',
    author_email='leonardbinet@gmail.com',
    url='https://github.com/leonardbinet/navitia-python',
    classifiers=[],
)
